/*  g_sysv_queue.c - SystemV IPC QUEUES TEST
 *  
 *  (C) Copyright XII 2004 Michal 'GiM' Spadlinski <gim at skrzynka dot pl>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License Version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "g_sysv_queue.h"
#include "dynstuff.h"
#include "g_ipc_wspolne.h"

extern const int g_data_format_s;
extern struct g_data_format parentp[2], childp[2];
int showtimes(char *str, list_t p);

int child_sysv_queue(int off)
{
	int queue;
	char *p;
	int i, len;
	struct timeval s, k;
	list_t sending = NULL, receiving = NULL;

	for (i=0; i<MAX_OPEN_TRIES; i++)
	{
		if ((queue = msgget(G_KEY, 0666)) != -1)
			break;
		SLEEP(s, 500);
	}

	if (i == MAX_OPEN_TRIES)
	{
		perror ("child: creating queue");
		return -1;
	}
	DEB("[%d] child queue opened %d\n", getpid(), queue);

	childp[off+0].type = 1;
	childp[off+0].id = getpid();
	p = childp[off+0].p;
	for (i=0; i<IPCQ_S; i++)
		*p++ = 'A';

	DEB("[%d] child queue data prepared\n", getpid());

	for (i=0; i<MAX_QUERIES; i++)
	{
		TIMESTAMP(s);
		DEB("[%d] child sending data\n", getpid());
		msgsnd(queue, &(childp[off+0]), g_data_format_s, 0);		
		DEB("[%d] child data sent\n", getpid());
		TIMESTAMP(k);

		DIFFTIME(k, s);
		list_add(&sending, &k, sizeof(struct timeval));

		TIMESTAMP(s);
		len = msgrcv(queue, &(childp[off+1]), g_data_format_s, getpid(), 0);
		TIMESTAMP(k);
		DIFFTIME(k, s);
		list_add(&receiving, &k, sizeof(struct timeval));

		childp[off+1].p[len - 2*long_s] = 0; 
		DEB("[%d] child received data from %li\n", getpid(), childp[off+1].id);
	}
	showtimes("[+] sysv msg queue   sending", sending);
	showtimes("[+] sysv msg queue receiving", receiving);


	list_destroy(sending, 1);
	list_destroy(receiving, 1);

	return 0;
}

int parent_sysv_queue(int pid)
{
	int queue, len;
	struct msqid_ds p;

	if ((queue = msgget(G_KEY, IPC_CREAT|0666|IPC_EXCL)) == -1)
	{
		perror ("parent creating queue");
		return -1;
	}
	DEB("[%d] parent queue created %d\n", getpid(), queue);

	if (msgctl(queue, IPC_STAT, &p) == -1)
	{
		perror("parent stat queue");
		return -1;
	}
	DEB("[%d] parent queue stated\n", getpid());

	while (1)
	{
		DEB("[%d] parent queue waiting for data\n", getpid());
		if ((len=msgrcv(queue, &(parentp[0]), g_data_format_s, 1, IPC_NOWAIT)) == -1)
		{
			if (waitpid(pid, NULL, WNOHANG) != 0)
				break;
		} else {
			DEB("[%d] parent queue data received from %li, %d byte[s]\n", getpid(), parentp[0].id, len);
			parentp[0].p[len] = '\0';
			parentp[0].type = parentp[0].id;
			parentp[0].id = getpid();
			msgsnd(queue, &(parentp[0]), g_data_format_s, 0);
		}
	}

	DEB("[%d] parent child exited\n", getpid());
	DEB("%li %li\n", p.msg_qnum, p.msg_qbytes);

	if (msgctl(queue, IPC_RMID, NULL) == -1)
	{
		perror("parent removing queue");
		return -1;
	}

	return 0;
}
